package control;

import javafx.scene.shape.Rectangle;
import model.ManModel;
import model.Model;
import view.View;

import java.util.*;

public class Controller {

    private static Random loto = new Random(Calendar.getInstance().getTimeInMillis());
    protected Model model;
    protected View view;
    protected ControllerKeyboard keyboard;
    protected ControllerAnimation animator;

    public Controller(Model model, View view) {
        this.model = model;
        this.view = view;
        keyboard = new ControllerKeyboard(model,view,this);
        animator = new ControllerAnimation(model, view, this);

    }

    public void startAnimation() {
        animator.start();
    }
    public void stopAnimation() {
        animator.stop();
    }

    public void step() {
        // If abnormal behavior is observed, uncomment next line to check abnormal collision case
        //checkAbnormalCollision();

        checkFloor();
        /* NB: in that example, since there are no other moving objects,
            if the Man is idle and no left/right key is pressed, then
            no need to do a collision test. In any other cases, do it
         */
        int manState = model.getMan().getState();
        if ( (model.isKeyPressed()) || (manState != ManModel.STATE_IDLE)) {
            // if still idle, determines new state : moving or jumping
            if (manState == ManModel.STATE_IDLE) {
                String lastKey = model.getLastKeyPressed();
                if ("Space".equals(lastKey)) {
                    int idxLeft = model.indexKeyPressed("Left");
                    int idxRight = model.indexKeyPressed("Right");
                    if ((idxLeft == -1) && (idxRight == -1)) {
                        model.getMan().setStateJump(ManModel.JUMP_UP);
                    } else if (idxLeft > idxRight) {
                        model.getMan().setStateJump(ManModel.JUMP_LEFT);
                    } else if (idxLeft <= idxRight) {
                        model.getMan().setStateJump(ManModel.JUMP_RIGHT);
                    }
                }
                else if ("Left".equals(lastKey)) {
                    model.getMan().setStateMoveLeft();
                }
                else if ("Right".equals(lastKey)) {
                    model.getMan().setStateMoveRight();
                }
            }
            // now check collisions
            checkCollisions();
        }
        model.update();
        view.update();
        model.resetFlags();
    }

    /* NB: this method is norammly used only for debug
     */
    public void checkAbnormalCollision() {

        // check abnormal case: man is already in collision in its current position
        boolean collide = false;
        for (Rectangle o : view.getObstacles()) {
            if (view.getManLook().checkCollision(o)) {
                collide = true;
                break;
            }
        }
        if (collide) {
            System.out.println("ABNORMAL: collision at current position");
        }
    }

    public void checkFloor() {
        // get current position
        double[] srcPos = model.getMan().getPosition();
        // first, if not already jumping/falling, check if there is a floor below the man
        if (model.getMan().getState() == ManModel.STATE_IDLE) {
            view.getManLook().setHitBoxPosition(srcPos[0], srcPos[1] + 1);
            int quadrants = 0;
            for (Rectangle o : view.getObstacles()) {
                int q = view.getManLook().checkQuadrantCollision(o);
                quadrants = quadrants | q;
            }
            if (quadrants == 0) {
                // lets fall
                System.out.println("no floor => falling");
                model.getMan().setStateFall();
            } else if (quadrants == 4) {
                // only bottom-left of Man is on the floor => lets move a bit the man to the right to slide
                System.out.println("just bottom left => slide right");
                srcPos[0] += 2;
                model.getMan().put(srcPos[0], srcPos[1]);
                model.getMan().setStateFall();
            } else if (quadrants == 8) {
                // only bottom-right of Man is on the floor => lets move a bit the man to the left to slide
                System.out.println("just bottom right => slide left");
                srcPos[0] -= 2;
                model.getMan().put(srcPos[0], srcPos[1]);
                model.getMan().setStateFall();
            }
        }
    }

    public void checkCollisions() {

        boolean collide = false;
        // get current & next position
        double[] srcPos = model.getMan().getPosition();
        double[] destPos = model.getMan().getNextPosition();

        double goodX = 0.0; // final "good" x position, i.e. without a collision
        double goodY = 0.0;  // final "good" y position, i.e. without a collision

        // get positions along the path defined by srcPos & destPos (pixelPath is in already in reversed order, see bresenham() )
        List<int[]> pixelPath = bresenham((int)srcPos[0], (int)srcPos[1], (int)destPos[0], (int)destPos[1]);
        // now follow the reversed path until no collision is detected.
        boolean noCollisionsAtDest = false;
        for (int i = 0;i<pixelPath.size();i++) {
            int[] t = pixelPath.get(i);
            view.getManLook().setHitBoxPosition(t[0], t[1]);
            collide = false;
            for (Rectangle o : view.getObstacles()) {
                if (view.getManLook().checkCollision(o)) {
                    collide = true;
                    break;
                }
            }
            if (!collide) {
                // find the good position
                goodX = t[0];
                goodY = t[1];
                if (i == 0) {
                    // no collision at dest => no collision at all
                    noCollisionsAtDest = true;
                }
                break; // go out of main loop
            }
        }
        // if collide is still true, no good position has been found => abnormal case !!
        if (collide) {
            System.out.print("CAUTION: no good position for a move from ");
            System.out.println(srcPos[0]+","+srcPos[1]+" to "+destPos[0]+","+destPos[1]);
            System.out.print("State = "+model.getMan().getState()+" Followed path: ");
            for (int i = 0;i<pixelPath.size(); i++) {
                int[] t = pixelPath.get(i);
                System.out.print(t[0] + "," + t[1] + " | ");
            }
            System.out.println();
        }
        if (!noCollisionsAtDest) {
            System.out.println("collide in state "+model.getMan().getState()+" must put in "+goodX+","+goodY);
            model.getMan().setStateCollision(goodX, goodY);
        }
        // else : achieve a "full" displacement, which is done in ManModel.update() by calling appropriate methods according to the state.
    }

    // CAUTION : this version of bresenham returns a list of points from x2,y2 to x1,y1
    private List<int[]> bresenham(int x1, int y1, int x2, int y2) {
        if (Math.abs(y2-y1) < Math.abs(x2-x1)) {
            if (x1 < x2) {
                return bresenhamLow(x1, y1, x2, y2, true);
            }
            else {
                return bresenhamLow(x2, y2, x1, y1, false);
            }
        }
        else {
            if (y1 < y2) {
                return bresenhamHigh(x1, y1, x2, y2, true);
            }
            else {
                return bresenhamHigh(x2, y2, x1, y1, false);
            }
        }
    }

    private List<int[]> bresenhamLow(int x1, int y1, int x2, int y2, boolean reversed) {
        List<int[]> lst = new ArrayList<>();
        int dx = x2 - x1;
        int dy = y2 - y1;
        int yi = 1;
        if (dy < 0) {
            yi = -1;
            dy = -dy;
        }

        int D = (2 * dy) - dx;
        int y = y1;

        for (int x = x1; x<=x2;x++) {
            int[] tab = new int[2];
            tab[0] = x;
            tab[1] = y;
            if (reversed) {
                lst.add(0,tab);
            }
            else {
                lst.add(tab);
            }
            if (D > 0 ) {
                y = y + yi;
                D = D + (2 * (dy - dx));
            }
            else {
                D = D + 2*dy;
            }
        }
        return lst;
    }
    private List<int[]> bresenhamHigh(int x1, int y1, int x2, int y2, boolean reversed) {
        List<int[]> lst = new ArrayList<>();
        int dx = x2 - x1;
        int dy = y2 - y1;
        int xi = 1;
        if (dx < 0) {
            xi = -1;
            dx = -dx;
        }

        int D = (2 * dx) - dy;
        int x = x1;

        for (int y = y1; y<=y2;y++) {
            int[] tab = new int[2];
            tab[0] = x;
            tab[1] = y;
            if (reversed) {
                lst.add(0,tab);
            }
            else {
                lst.add(tab);
            }
            if (D > 0 ) {
                x = x + xi;
                D = D + (2 * (dx - dy));
            }
            else {
                D = D + 2*dx;
            }
        }
        return lst;
    }
}
